#!/usr/bin/env python3
import click
import os
import json
import logging
from .project import project
from .cluster import cluster
from .profile import profile
from .cache import cache

from .utils import login_profile, install_completions

# Main CLI entry point
@click.group(invoke_without_command=True)
@click.option("--profile")
@click.option('--project-name', '-p', required = False, help="Project Name")
@click.option('-v', '--verbose', count=True)
@click.pass_context
def cli(ctx, project_name, profile, verbose):
    """
    CLI tool to manage projects and clusters.

    You can run commands on two scopes: 'project' and 'cluster'.
    Each scope has several actions like 'list', 'get', 'create', etc.
    """

    if not profile:
        profile = "default"

    if ctx.invoked_subcommand not in ['profile', 'fullhelp', 'version', 'install-completion', None]:
        login_profile(profile)

    LOGLEVEL = os.environ.get('LOGLEVEL', 'WARNING').upper()

    if verbose >= 1:
        LOGLEVEL = "INFO"

    if verbose >= 2:
        LOGLEVEL = "DEBUG"

    logging.basicConfig(
        format='%(asctime)s %(levelname)-8s %(filename)s:%(funcName)s %(message)s',
        level=LOGLEVEL,
        datefmt='%Y-%m-%d %H:%M:%S'
    )
    logging.info("Set loglevel to %s", LOGLEVEL)

    if ctx.invoked_subcommand is None:
        click.echo(ctx.get_help())

    if not hasattr(ctx, 'obj') or not ctx.obj:
        ctx.obj = dict()

    if project_name != None:
        ctx.obj['project_name'] = project_name

# Register the project and cluster groups with the main CLI
cli.add_command(project)
cli.add_command(cluster)
cli.add_command(profile)
cli.add_command(cache)

def recursive_help(cmd, parent=None):
    ctx = click.core.Context(cmd, info_name=cmd.name, parent=parent)
    print(cmd.get_help(ctx))
    print()
    commands = getattr(cmd, 'commands', {})
    for sub in commands.values():
        recursive_help(sub, ctx)

@cli.command('fullhelp', help="Display detailed help information for all commands.")
def fullhelp():
    recursive_help(cli)

@cli.command('version', help="Show the current CLI version.")
def version():
    import importlib.metadata
    print(importlib.metadata.version(__package__))

@cli.command("install-completion", help="Install shell completion scripts.")
@click.option('--type', help="Shell")
def install_completion(type):
    install_completions(type)

if __name__ == '__main__':
    cli()
