import click
import time
import datetime
import dateutil.parser
import human_readable
import prettytable
import os

from .utils import do_request, print_output, find_project_id_by_name, get_project_id, set_project_id, detect_and_parse_input, transform_tuple, ctx_update, set_cluster_id, get_template, get_project_name, format_changed_row, is_interesting_status

# DEIFNE THE PROJECT COMMAND GROUP
@click.group(help="Project related commands.")
@click.option('--project', 'project_name', required = False, help="Project Name")
@click.option('--project-name', '-p', required = False, help="Project Name")
@click.pass_context
def project(ctx, project_name):
    ctx_update(ctx, project_name)

# LOGIN ON PROJECT
@project.command('login', help="Set default project")
@click.option('--project-name', '-p', required=True, help="Name of project", type=click.STRING)
def project_login(project_name):
    data = do_request("GET", 'projects', params={"name": project_name})
    if len(data) != 1:
        raise click.BadParameter(
            f"{len(data)} projects found by name: {project_name}")
    project = data.pop()

    project_id = project['id']
    project_name = project['name']

    set_project_id(project_id)
    set_cluster_id("")

    project_name = click.style(project_name, bold=True)

    click.echo(f"Logged into project: {project_name}")

# LOGOUT ON PROJECT
@project.command('logout', help="Unset default project")
def project_logout():
    set_project_id("")
    set_cluster_id("")
    click.echo("Logged out from the current project")

# LIST PROJECTS
@project.command('list', help="List all projects")
@click.option('--project-name', '-p', help="Name of project", type=click.STRING)
@click.option('--deleted', is_flag=True, help="List deleted projects")
@click.option('--plain', is_flag=True, help="Plain table format")
@click.option('--msword', is_flag=True, help="Microsoft Word table format")
@click.option('--uuid', is_flag=True, help="show uuid")
@click.option('--watch', '-w', is_flag=True, help="Watch the changes")
@click.pass_context
def project_list(ctx, project_name, deleted, plain, msword, uuid, watch):
    project_name, _ = ctx_update(ctx, project_name)

    project_id = get_project_id()
    params = {}
    if project_name:
        params['name'] = project_name

    if deleted:
        params['deleted'] = True

    field_names = ["NAME", "CREATED", "UPDATED", "STATUS", "DEFAULT"]
    if uuid:
        field_names.append('UUID')

    table = prettytable.PrettyTable()
    table.field_names = field_names

    table._min_width = {"CREATED": 13, "UPDATED": 13, "STATUS": 10}

    if plain or watch:
        table.set_style(prettytable.PLAIN_COLUMNS)

    if msword:
        table.set_style(prettytable.MSWORD_FRIENDLY)

    def format_row(project):
        status = project.get('status')
        is_default = True if project.get('id') == project_id else False

        if status == 'ready':
            msg = click.style(status, fg='green')
        elif status == 'failed' or status == 'deleted':
            msg = click.style(status, fg='red')
        elif status == 'deploying':
            msg = click.style(status, fg='yellow')
        else:
            msg = status

        name = click.style(project['name'], bold=True)
        if is_default:
            default = "*"
        else:
            default = ""

        created_at = dateutil.parser.parse(project['created_at'])
        updated_at = dateutil.parser.parse(project['updated_at'])
        now = datetime.datetime.now(tz=created_at.tzinfo)

        row = [name, human_readable.date_time(now - created_at), human_readable.date_time(now - updated_at), msg, default]
        if uuid:
            row.append(project['id'])

        return row, status, project['name']

    data = do_request("GET", 'projects', params=params)
    initial_projects = {}

    for project in data:
        row, _, name = format_row(project)
        table.add_row(row)
        initial_projects[name] = project

    click.echo(table)


    if watch:
        total_sleep = 0
        try:
            while True:
                time.sleep(2)
                total_sleep += 2

                data = do_request("GET", 'projects', params=params)

                current_project_names = {project['name'] for project in data}

                for name, project in list(initial_projects.items()):
                    if name not in current_project_names:
                        deleted_project = project.copy()
                        deleted_project['status'] = 'deleted'

                        row, current_status, _ = format_row(deleted_project)

                        new_table = format_changed_row(table, row)
                        click.echo(new_table)

                        del initial_projects[name]

                for project in data:
                    row, current_status, name = format_row(project)

                    if name not in initial_projects:
                        new_table = format_changed_row(table, row)
                        click.echo(new_table)
                        initial_projects[name] = project
                        continue

                    stored_project = initial_projects[name]
                    project_status = stored_project.get('status')
                    if project_status != current_status:
                        new_table = format_changed_row(table, row)
                        click.echo(new_table)
                        initial_projects[name] = project
                        continue

                    if total_sleep % 10 == 0 and is_interesting_status(current_status):
                        new_table = format_changed_row(table, row)
                        click.echo(new_table)
                        initial_projects[name] = project



        except KeyboardInterrupt:
            click.echo("\nWatch stopped.")

# CREATE PROJECT BY NAME
@project.command('create', help="Create a new project")
@click.option('--project-name', '-p', help="Name of the project")
@click.option('--description', help="Description of the project")
@click.option('--cidr', help='CIDR for the project')
@click.option('--quirk', multiple=True, help="Quirk")
@click.option('--tags', help="Comma-separated list of tags, example: 'key1=value1,key2=value2'")
@click.option('--disable-api-termination', type=click.BOOL, help="Disable delete action by API")
@click.option('--dry-run', is_flag=True, help="Client dry-run, only print the object that would be sent, without sending it")
@click.option('-o', '--output', type=click.Choice(["json", "yaml", "silent"]), help="Specify output format, by default is json")
@click.option('-f', '--filename', type=click.File("r"), help="Path to file to use to create the project ")
@click.pass_context
def project_create(ctx, project_name, description, cidr, quirk, tags, disable_api_termination, dry_run, output, filename):
    project_name, _ = ctx_update(ctx, project_name)
    project_config = None

    if filename:
        input_data = filename.read()
        project_config = detect_and_parse_input(input_data)
    else:
        project_config = get_template("project")
        if os.getenv("OKS_REGION"):
            project_config["region"] = os.getenv("OKS_REGION")

    if project_name:
        project_config['name'] = project_name

    if description:
        project_config['description'] = description

    if cidr:
        project_config['cidr'] = cidr

    if quirk:
        project_config['quirks'] = transform_tuple(quirk)
    
    if tags:
        parsed_tags = {}

        pairs = tags.split(',')
        for pair in pairs:
            if '=' not in pair:
                raise click.ClickException(f"Malformed tags: '{pair}' (expected key=value)")
            key, value = pair.split('=', 1)
            parsed_tags[key.strip()] = value.strip()

        project_config['tags'] = parsed_tags

    if disable_api_termination is not None:
        project_config["disable_api_termination"] = disable_api_termination

    if not dry_run:
        data = do_request("POST", 'projects', json=project_config)
        print_output(data, output)
    else:
        print_output(project_config, output)

# GET PROJECT BY NAME
@project.command('get', help="Get default project or the project by name")
@click.option('--project-name', '-p', help="Name of the project")
@click.option('-o', '--output', type=click.Choice(["json", "yaml"]), help="Specify output format, by default is json")
@click.pass_context
def project_get(ctx, project_name, output):
    project_name, _ = ctx_update(ctx, project_name)

    project_id = find_project_id_by_name(project_name)

    data = do_request("GET", f'projects/{project_id}')
    print_output(data, output)

# DELETE PROJECT BY NAME
@project.command('delete', help="Delete a project by ID")
@click.option('--project-name', '-p', required=False, help="Project Name")
@click.option('-o', '--output', type=click.Choice(["json", "yaml"]), help="Specify output format, by default is json")
@click.option('--dry-run', is_flag=True, help="Run without any action")
@click.option('--force', is_flag=True, help="Force deletion without confirmation")
@click.pass_context
def project_delete_command(ctx, project_name, output, dry_run, force):
    project_name, _ = ctx_update(ctx, project_name)

    project_id = find_project_id_by_name(project_name)
    project_name = get_project_name(project_name)

    current_project_id = get_project_id()

    if dry_run:
        click.echo("Dry run: The project would be deleted.")
    else:
        if force or click.confirm(f"Are you sure you want to delete the project with name {project_name}?", abort=True):
            data = do_request("DELETE", f'projects/{project_id}')

            if current_project_id == project_id:
                set_project_id("")
                set_cluster_id("")
            print_output(data, output)


# UPDATE PROJECT BY NAME
@project.command('update', help="Update a project by ID")
@click.option('--project-name', '-p', required=False, help="Project Name")
@click.option('--quirk', multiple=True, help="Quirk")
@click.option('--tags', help="Comma-separated list of tags, example: 'key1=value1,key2=value2'")
@click.option('--disable-api-termination', type=click.BOOL, help="Disable delete action by API")
@click.option('-o', '--output', type=click.Choice(["json", "yaml"]), help="Specify output format, by default is json")
@click.option('--dry-run', is_flag=True, help="Run without any action")
@click.pass_context
def project_update_command(ctx, project_name, quirk, tags, disable_api_termination, output, dry_run):
    project_name, _ = ctx_update(ctx, project_name)

    project_id = find_project_id_by_name(project_name)

    project_config = {}

    if disable_api_termination is not None:
        project_config["disable_api_termination"] = disable_api_termination

    if quirk:
        project_config['quirks'] = transform_tuple(quirk)

    if tags is not None:
        parsed_tags = {}

        if not len(tags) == 0:
            pairs = tags.split(',')
            for pair in pairs:
                if '=' not in pair:
                    raise click.ClickException(f"Malformed tags: '{pair}' (expected key=value)")
                key, value = pair.split('=', 1)
                parsed_tags[key.strip()] = value.strip()

        project_config['tags'] = parsed_tags

    if dry_run:
        print_output(project_config, output)
    else:
        data = do_request("PATCH", f'projects/{project_id}', json = project_config)
        print_output(data, output)

# GET PROJECT QUOTAS BY PROJECT NAME
@project.command('quotas', help="Get project quotas")
@click.option('--project-name', '-p', help="Name of the project")
@click.option('-o', '--output', type=click.Choice(["json", "yaml"]), help="Specify output format, by default is json")
@click.pass_context
def project_get(ctx, project_name, output):
    project_name, _ = ctx_update(ctx, project_name)

    project_id = find_project_id_by_name(project_name)

    data = do_request("GET", f'projects/{project_id}/quotas')["data"]["quotas"]
    print_output(data, output)
